# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from kolena._api.v1.batched_load import BatchedLoad
from kolena._utils.pydantic_v1.dataclasses import dataclass


class Model:
    class Path(str, Enum):
        CREATE = "/generic/model/create"
        LOAD = "/generic/model/load"
        LOAD_ALL = "/generic/model/load-all"
        DELETE = "/generic/model/delete"
        LOAD_INFERENCES = "/generic/model/load-inferences"

    @dataclass(frozen=True)
    class LoadInferencesRequest(BatchedLoad.BaseInitDownloadRequest):
        model_id: int
        test_case_id: int


class TestCase:
    class Path(str, Enum):
        CREATE = "/generic/test-case/create"
        LOAD = "/generic/test-case/load"
        INIT_LOAD_TEST_SAMPLES = "/generic/test-case/load-test-samples"
        COMPLETE_EDIT = "/generic/test-case/edit"
        BULK_PROCESS = "/generic/test-case/bulk-process"


class TestSuite:
    class Path(str, Enum):
        CREATE = "/generic/test-suite/create"
        LOAD = "/generic/test-suite/load"
        LOAD_ALL = "/generic/test-suite/load-all"
        EDIT = "/generic/test-suite/edit"
        DELETE = "/generic/test-suite/delete"
        INIT_LOAD_TEST_SAMPLES = "/generic/test-suite/load-test-samples"

    @dataclass(frozen=True)
    class LoadTestSamplesRequest(BatchedLoad.BaseInitDownloadRequest):
        test_suite_id: int


class TestRun:
    class Path(str, Enum):
        CREATE_OR_RETRIEVE = "/generic/test-run/create-or-retrieve"
        MARK_CRASHED = "/generic/test-run/mark-crashed"
        EVALUATE = "/generic/test-run/evaluate"
        LOAD_TEST_SAMPLES = "/generic/test-run/load-remaining-test-samples"
        LOAD_INFERENCES = "/generic/test-run/load-test-sample-inferences"
        UPLOAD_INFERENCES = "/generic/test-run/upload-inferences"
        UPLOAD_TEST_SAMPLE_METRICS = "/generic/test-run/upload-test-sample-metrics"
        UPLOAD_TEST_SAMPLE_METRICS_THRESHOLDED = "/generic/test-run/upload-test-sample-thresholded-metrics"
        UPLOAD_TEST_CASE_METRICS = "/generic/test-run/upload-test-case-metrics"
        UPLOAD_TEST_CASE_PLOTS = "/generic/test-run/upload-test-case-plots"
        UPLOAD_TEST_SUITE_METRICS = "/generic/test-run/upload-test-suite-metrics"
        UPDATE_METRICS_STATUS = "/generic/test-run/update-metrics-status"

    @dataclass(frozen=True)
    class EvaluatorConfiguration:
        display_name: str
        configuration: Dict[str, Any]  # TODO: real type is JSON

    @dataclass(frozen=True)
    class CreateOrRetrieveRequest:
        model_id: int
        test_suite_id: int
        evaluator: Optional[str] = None
        configurations: Optional[List["TestRun.EvaluatorConfiguration"]] = None

    @dataclass(frozen=True)
    class CreateOrRetrieveResponse:
        test_run_id: int

    @dataclass(frozen=True)
    class EvaluateRequest:
        test_run_id: int

    EvaluateResponse = CreateOrRetrieveResponse

    @dataclass(frozen=True)
    class LoadRemainingTestSamplesRequest(BatchedLoad.BaseInitDownloadRequest):
        test_run_id: int
        load_all: bool = False

    @dataclass(frozen=True)
    class LoadTestSampleInferencesRequest(BatchedLoad.BaseInitDownloadRequest):
        test_run_id: int

    @dataclass(frozen=True)
    class UploadInferencesRequest(BatchedLoad.WithLoadUUID):
        test_run_id: int
        reset: bool = False

    @dataclass(frozen=True)
    class UploadTestSampleMetricsRequest(BatchedLoad.WithLoadUUID):
        test_run_id: int
        test_case_id: Optional[int]
        configuration: Optional["TestRun.EvaluatorConfiguration"]

    @dataclass(frozen=True)
    class UploadTestSampleThresholdedMetricsRequest(BatchedLoad.WithLoadUUID):
        test_run_id: int
        test_case_id: Optional[int]
        configuration: Optional["TestRun.EvaluatorConfiguration"]
        model_id: int

    @dataclass(frozen=True)
    class UpdateMetricsStatusRequest:
        test_run_id: int
        progress: float
        message: str = ""

    @dataclass(frozen=True)
    class UploadAggregateMetricsRequest(BatchedLoad.WithLoadUUID):
        test_run_id: int
        test_suite_id: int

    @dataclass(frozen=True)
    class BulkUploadResponse:
        n_uploaded: int


class Workflow:
    class Path(str, Enum):
        EVALUATOR = "/generic/workflow/evaluator"

    @dataclass(frozen=True)
    class EvaluatorRoleConfig:
        job_role_name: str
        job_role_arn: str
        external_id: str
        assume_role_arn: str

    @dataclass(frozen=True)
    class RegisterEvaluatorRequest:
        workflow: str
        name: str
        image: str
        secret: Optional[str] = None
        aws_assume_role: Optional[str] = None

    @dataclass(frozen=True)
    class EvaluatorResponse:
        name: str
        image: Optional[str] = None
        created: Optional[str] = None
        secret: Optional[str] = None
        aws_role_config: Optional["Workflow.EvaluatorRoleConfig"] = None

    @dataclass(frozen=True)
    class ListEvaluatorsResponse:
        evaluators: List["Workflow.EvaluatorResponse"]


class Search:
    class Path(str, Enum):
        EMBEDDINGS = "/generic/search/embeddings"

    @dataclass(frozen=True)
    class UploadEmbeddingsRequest(BatchedLoad.WithLoadUUID):
        ...

    @dataclass(frozen=True)
    class UploadEmbeddingsResponse:
        n_samples: int


TestRun.CreateOrRetrieveRequest.__pydantic_model__.update_forward_refs()  # type: ignore
TestRun.UploadTestSampleMetricsRequest.__pydantic_model__.update_forward_refs()  # type: ignore
TestRun.UploadTestSampleThresholdedMetricsRequest.__pydantic_model__.update_forward_refs()  # type: ignore
Workflow.EvaluatorResponse.__pydantic_model__.update_forward_refs()  # type: ignore
Workflow.ListEvaluatorsResponse.__pydantic_model__.update_forward_refs()  # type: ignore



# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Annotations are visualized in Kolena as overlays on top of datapoints.

| Annotation | Conditions for visualization |
| --- | --- |
| [`BoundingBox`][kolena.annotation.BoundingBox] | Limited to `Image` or `Video` data |
| [`BoundingBox3D`][kolena.annotation.BoundingBox3D] | Limited to `PointCloud` data |
| [`Polygon`][kolena.annotation.Polygon] | Limited to `Image` or `Video` data |
| [`Polyline`][kolena.annotation.Polyline] | Limited to `Image` or `Video` data |
| [`Keypoints`][kolena.annotation.Keypoints] | Limited to `Image` or `Video` data |
| [`SegmentationMask`][kolena.annotation.SegmentationMask] | Limited to `Image` or `Video` data |
| [`BitmapMask`][kolena.annotation.BitmapMask] | Limited to `Image` or `Video` data |
| [`Label`][kolena.annotation.Label] | Valid for all data |
| [`TimeSegment`][kolena.annotation.TimeSegment] | Limited to `Audio` or `Video` data |
| [`TextSegment`][kolena.annotation.TextSegment] | Limited to `Text` data |

For example, when viewing images in the Studio, any annotations (such as lists of
[`BoundingBox`][kolena.annotation.BoundingBox] objects) present in the datapoints are
rendered on top of the image.
"""  # noqa: E501
import dataclasses
from abc import ABCMeta
from functools import reduce
from typing import Dict
from typing import List
from typing import Tuple

from kolena._utils.datatypes import DataCategory
from kolena._utils.datatypes import DataType
from kolena._utils.datatypes import TypedDataObject
from kolena._utils.pydantic_v1.dataclasses import dataclass
from kolena._utils.validators import ValidatorConfig


class _AnnotationType(DataType):
    BOUNDING_BOX = "BOUNDING_BOX"
    POLYGON = "POLYGON"
    POLYLINE = "POLYLINE"
    POLYLINE_3D = "POLYLINE_3D"
    KEYPOINTS = "KEYPOINTS"
    KEYPOINTS_3D = "KEYPOINTS_3D"
    BOUNDING_BOX_3D = "BOUNDING_BOX_3D"
    SEGMENTATION_MASK = "SEGMENTATION_MASK"
    BITMAP_MASK = "BITMAP_MASK"
    LABEL = "LABEL"
    TIME_SEGMENT = "TIME_SEGMENT"
    TEXT_SEGMENT = "TEXT_SEGMENT"

    @staticmethod
    def _data_category() -> DataCategory:
        return DataCategory.ANNOTATION


@dataclass(frozen=True, config=ValidatorConfig)
class Annotation(TypedDataObject[_AnnotationType], metaclass=ABCMeta):
    """The base class for all annotation types."""


@dataclass(frozen=True, config=ValidatorConfig)
class BoundingBox(Annotation):
    """
    Rectangular bounding box specified with pixel coordinates of the top left and bottom right vertices.

    The reserved fields `width`, `height`, `area`, and `aspect_ratio` are automatically populated with values derived
    from the provided coordinates.
    """

    top_left: Tuple[float, float]
    """The top left vertex (in `(x, y)` pixel coordinates) of this bounding box."""

    bottom_right: Tuple[float, float]
    """The bottom right vertex (in `(x, y)` pixel coordinates) of this bounding box."""

    width: float = dataclasses.field(init=False)
    height: float = dataclasses.field(init=False)
    area: float = dataclasses.field(init=False)
    aspect_ratio: float = dataclasses.field(init=False)

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.BOUNDING_BOX

    def __post_init__(self) -> None:
        object.__setattr__(self, "width", abs(self.bottom_right[0] - self.top_left[0]))
        object.__setattr__(self, "height", abs(self.bottom_right[1] - self.top_left[1]))
        object.__setattr__(self, "area", self.width * self.height)
        object.__setattr__(self, "aspect_ratio", self.width / self.height if self.height != 0 else 0)


@dataclass(frozen=True, config=ValidatorConfig)
class LabeledBoundingBox(BoundingBox):
    """
    Rectangular bounding box specified with pixel coordinates of the top left and bottom right vertices and a string
    label.
    """

    label: str
    """The label (e.g. model classification) associated with this bounding box."""


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredBoundingBox(BoundingBox):
    """
    Rectangular bounding box specified with pixel coordinates of the top left and bottom right vertices and a float
    score.
    """

    score: float
    """The score (e.g. model confidence) associated with this bounding box."""


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredLabeledBoundingBox(BoundingBox):
    """
    Rectangular bounding box specified with pixel coordinates of the top left and bottom right vertices, a string
    label, and a float score.
    """

    label: str
    """The label (e.g. model classification) associated with this bounding box."""

    score: float
    """The score (e.g. model confidence) associated with this bounding box."""


@dataclass(frozen=True, config=ValidatorConfig)
class Polygon(Annotation):
    """Arbitrary polygon specified by three or more pixel coordinates."""

    points: List[Tuple[float, float]]
    """The sequence of `(x, y)` pixel coordinates comprising the boundary of this polygon."""

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.POLYGON

    def __post_init__(self) -> None:
        if len(self.points) < 3:
            raise ValueError(f"{type(self).__name__} must have at least three points ({len(self.points)} provided)")


@dataclass(frozen=True, config=ValidatorConfig)
class LabeledPolygon(Polygon):
    """Arbitrary polygon specified by three or more pixel coordinates and a string label."""

    label: str
    """The label (e.g. model classification) associated with this polygon."""


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredPolygon(Polygon):
    """
    Arbitrary polygon specified by three or more pixel coordinates and a float score.
    """

    score: float
    """The score (e.g. model confidence) associated with this polygon."""


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredLabeledPolygon(Polygon):
    """
    Arbitrary polygon specified by three or more pixel coordinates with a string label and a float score.
    """

    label: str
    """The label (e.g. model classification) associated with this polygon."""

    score: float
    """The score (e.g. model confidence) associated with this polygon."""


@dataclass(frozen=True, config=ValidatorConfig)
class Keypoints(Annotation):
    """Array of any number of keypoints specified in pixel coordinates."""

    points: List[Tuple[float, float]]
    """The sequence of discrete `(x, y)` pixel coordinates comprising this keypoints annotation."""

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.KEYPOINTS


@dataclass(frozen=True, config=ValidatorConfig)
class Keypoints3D(Annotation):
    """Array of any number of keypoints specified in a right-handed coordinate system."""

    points: List[Tuple[float, float, float]]
    """The sequence of discrete `(x, y, z)` coordinates comprising this keypoints annotation."""

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.KEYPOINTS_3D


@dataclass(frozen=True, config=ValidatorConfig)
class Polyline(Annotation):
    """Polyline with any number of vertices specified in pixel coordinates."""

    points: List[Tuple[float, float]]
    """The sequence of connected `(x, y)` pixel coordinates comprising this polyline."""

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.POLYLINE


@dataclass(frozen=True, config=ValidatorConfig)
class Polyline3D(Annotation):
    """A three-dimensional Polyline with any number of vertices specified a right-handed coordinate system."""

    points: List[Tuple[float, float, float]]
    """The sequence of connected `(x, y, z)` coordinates comprising this polyline."""

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.POLYLINE_3D


@dataclass(frozen=True, config=ValidatorConfig)
class BoundingBox3D(Annotation):
    """
    Three-dimensional cuboid bounding box in a right-handed coordinate system.

    Specified by `(x, y, z)` coordinates for the `center` of the cuboid, `(x, y, z)` `dimensions`, and a `rotation`
    parameter specifying the degrees of rotation about each axis `(x, y, z)` ranging `[-π, π]`.

    The reserved field `volume` is automatically derived from the provided `dimensions`.
    """

    center: Tuple[float, float, float]
    """`(x, y, z)` coordinates specifying the center of the bounding box."""

    dimensions: Tuple[float, float, float]
    """`(x, y, z)` measurements specifying the dimensions of the bounding box."""

    rotations: Tuple[float, float, float]
    """Rotations in degrees about each `(x, y, z)` axis."""

    volume: float = dataclasses.field(init=False)

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.BOUNDING_BOX_3D

    def __post_init__(self) -> None:
        object.__setattr__(self, "volume", reduce(lambda a, b: a * b, self.dimensions))


@dataclass(frozen=True, config=ValidatorConfig)
class LabeledBoundingBox3D(BoundingBox3D):
    """[`BoundingBox3D`][kolena.workflow.annotation.BoundingBox3D] with an additional string label."""

    label: str
    """The label associated with this 3D bounding box."""


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredBoundingBox3D(BoundingBox3D):
    """[`BoundingBox3D`][kolena.workflow.annotation.BoundingBox3D] with an additional float score."""

    score: float
    """The score associated with this 3D bounding box."""


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredLabeledBoundingBox3D(BoundingBox3D):
    """[`BoundingBox3D`][kolena.workflow.annotation.BoundingBox3D] with an additional string label and float score."""

    label: str
    """The label associated with this 3D bounding box."""

    score: float
    """The score associated with this 3D bounding box."""


@dataclass(frozen=True, config=ValidatorConfig)
class SegmentationMask(Annotation):
    """
    Raster segmentation mask. The `locator` is the URL to the image file representing the segmentation mask.

    The segmentation mask must be rendered as a single-channel, 8-bit-depth (grayscale) image. For the best results,
    use a lossless file format such as PNG. Each pixel's value is the numerical ID of its class label, as specified in
    the `labels` map. Any pixel value not present in the `labels` map is rendered as part of the background.

    For example, `labels = {255: "object"}` will highlight all pixels with the value of 255 as `"object"`. Every
    other pixel value will be transparent.
    """

    labels: Dict[int, str]
    """Mapping of unique label IDs (pixel values) to unique label values."""

    locator: str
    """URL of the segmentation mask image."""

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.SEGMENTATION_MASK


@dataclass(frozen=True, config=ValidatorConfig)
class BitmapMask(Annotation):
    """Arbitrary bitmap mask. The `locator` is the URL to the image file representing the mask."""

    locator: str
    """URL of the bitmap data."""

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.BITMAP_MASK


@dataclass(frozen=True, config=ValidatorConfig)
class Label(Annotation):
    """Label, e.g. for classification."""

    label: str
    """String label for this classification."""

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.LABEL


ClassificationLabel = Label
"""Alias for [`Label`][kolena.workflow.annotation.Label]."""


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredLabel(Label):
    """Label with accompanying score."""

    score: float
    """Score associated with this label."""


ScoredClassificationLabel = ScoredLabel
"""Alias for [`ScoredLabel`][kolena.workflow.annotation.ScoredLabel]."""


@dataclass(frozen=True, config=ValidatorConfig)
class TimeSegment(Annotation):
    """
    Segment of time in the associated audio or video file.

    When a `group` is specified, segments are displayed on Kolena with different colors for each group present in a
    `List[TimeSegment]`. Example usage:

    ```py
    transcription: List[TimeSegment] = [
        LabeledTimeSegment(group="A", label="Knock, knock.", start=0, end=1),
        LabeledTimeSegment(group="B", label="Who's there?", start=2, end=3),
        LabeledTimeSegment(group="A", label="Example.", start=3.5, end=4),
        LabeledTimeSegment(group="B", label="Example who?", start=4.5, end=5.5),
        LabeledTimeSegment(group="A", label="Example illustrating two-person dialogue using `group`.", start=6, end=9),
    ]
    ```
    """

    start: float
    """Start time, in seconds, of this segment."""

    end: float
    """End time, in seconds, of this segment."""

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.TIME_SEGMENT


@dataclass(frozen=True, config=ValidatorConfig)
class LabeledTimeSegment(TimeSegment):
    """Time segment with accompanying label, e.g. audio transcription."""

    label: str
    """The label associated with this time segment."""


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredTimeSegment(TimeSegment):
    """Time segment with additional float score, representing e.g. model prediction confidence."""

    score: float
    """The score associated with this time segment."""


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredLabeledTimeSegment(TimeSegment):
    """Time segment with accompanying label and score."""

    label: str
    """The label associated with this time segment."""

    score: float
    """The score associated with this time segment."""


@dataclass(frozen=True, config=ValidatorConfig)
class TextSegment(Annotation):
    """
    Represents a segment of text within a specified text field to highlight.
    The `start` index is inclusive and the `end` index is
    exclusive, following the convention of Python string slicing.
    The following example adds highlights to two text fields
    named `text` and `summary`.

    ```py
    text = "Hello, world"
    summary = "A summary of current documentation."

    text_segments: List[TextSegment] = [
        # Highlights "Hello" in "Hello, world"
        TextSegment(text_field="text", start=0, end=5),
        # Highlights "summary of" in "A summary of current documentation."
        TextSegment(text_field="summary", start=2, end=12),
    ]
    ```
    """

    text_field: str
    """Text field column name containing the text segments."""

    start: int
    """Zero-indexed start index (inclusive) of the text segment."""

    end: int
    """Zero-indexed end index (exclusive) of the text segment."""

    def __post_init__(self) -> None:
        if self.start < 0:
            raise ValueError(f"Start index must be non-negative ({self.start} provided)")
        if self.end < 0:
            raise ValueError(f"End index must be non-negative ({self.end} provided)")
        if self.start >= self.end:
            raise ValueError(f"Start index must be less than end index ({self.start} >= {self.end})")

    @staticmethod
    def _data_type() -> _AnnotationType:
        return _AnnotationType.TEXT_SEGMENT


@dataclass(frozen=True, config=ValidatorConfig)
class LabeledTextSegment(TextSegment):
    """Text segment with accompanying label, e.g. Location."""

    label: str
    """The label associated with this text segment."""


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredTextSegment(TextSegment):
    """Text segment with additional float score, representing e.g. model prediction confidence."""

    score: float
    """The score associated with this text segment."""


@dataclass(frozen=True, config=ValidatorConfig)
class ScoredLabeledTextSegment(TextSegment):
    """Text segment with accompanying label and score."""

    label: str
    """The label associated with this text segment."""

    score: float
    """The score associated with this text segment."""


_ANNOTATION_TYPES = [
    BoundingBox,
    LabeledBoundingBox,
    ScoredBoundingBox,
    ScoredLabeledBoundingBox,
    Polygon,
    LabeledPolygon,
    ScoredPolygon,
    ScoredLabeledPolygon,
    Keypoints,
    Polyline,
    BoundingBox3D,
    LabeledBoundingBox3D,
    ScoredBoundingBox3D,
    ScoredLabeledBoundingBox3D,
    SegmentationMask,
    BitmapMask,
    Label,
    ScoredLabel,
    ClassificationLabel,
    ScoredClassificationLabel,
    TimeSegment,
    LabeledTimeSegment,
    ScoredTimeSegment,
    ScoredLabeledTimeSegment,
    TextSegment,
    LabeledTextSegment,
    ScoredTextSegment,
    ScoredLabeledTextSegment,
    Keypoints3D,
    Polyline3D,
]



# Copyright 2021-2024 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from dataclasses import asdict
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Tuple
from typing import Union

import pandas as pd

from kolena._api.v1.event import EventAPI
from kolena._api.v2.model import LoadResultsRequest
from kolena._api.v2.model import Path
from kolena._api.v2.model import UploadResultsRequest
from kolena._api.v2.model import UploadResultsResponse
from kolena._utils import krequests_v2 as krequests
from kolena._utils import log
from kolena._utils.batched_load import _BatchedLoader
from kolena._utils.batched_load import init_upload
from kolena._utils.batched_load import upload_data_frame
from kolena._utils.consts import BatchSize
from kolena._utils.endpoints import get_platform_url
from kolena._utils.endpoints import serialize_models_url
from kolena._utils.instrumentation import with_event
from kolena._utils.serde import from_dict
from kolena._utils.state import API_V2
from kolena.dataset._common import COL_DATAPOINT
from kolena.dataset._common import COL_DATAPOINT_ID_OBJECT
from kolena.dataset._common import COL_EVAL_CONFIG
from kolena.dataset._common import COL_RESULT
from kolena.dataset._common import COL_THRESHOLDED_OBJECT
from kolena.dataset._common import DEFAULT_SOURCES
from kolena.dataset._common import validate_batch_size
from kolena.dataset._common import validate_dataframe_columns
from kolena.dataset._common import validate_dataframe_ids
from kolena.dataset.dataset import _load_dataset_metadata
from kolena.dataset.dataset import _to_deserialized_dataframe
from kolena.dataset.dataset import _to_serialized_dataframe
from kolena.errors import IncorrectUsageError

EvalConfig = Optional[Dict[str, Any]]
"""
User defined configuration for evaluating results, for example `{"threshold": 7}`.
"""
DataFrame = Union[pd.DataFrame, Iterator[pd.DataFrame]]
"""
A type alias representing a DataFrame, which can be either a pandas DataFrame
or an iterator of pandas DataFrames.
"""


class EvalConfigResults(NamedTuple):
    """
    Named tuple where the first element (the `eval_config` field) is an evaluation configuration, and the second element
    (the `results` field) is the corresponding DataFrame of results.
    """

    eval_config: EvalConfig
    results: pd.DataFrame


def _iter_result_raw(
    dataset: str,
    model: str,
    batch_size: int,
    commit: Optional[str] = None,
    include_extracted_properties: bool = False,
) -> Iterator[pd.DataFrame]:
    validate_batch_size(batch_size)
    init_request = LoadResultsRequest(
        dataset=dataset,
        model=model,
        batch_size=batch_size,
        commit=commit,
        include_extracted_properties=include_extracted_properties,
    )
    yield from _BatchedLoader.iter_data(
        init_request=init_request,
        endpoint_path=Path.LOAD_RESULTS.value,
        df_class=None,
        endpoint_api_version=API_V2,
    )


def _fetch_results(
    dataset: str,
    model: str,
    commit: Optional[str] = None,
    include_extracted_properties: bool = False,
) -> pd.DataFrame:
    df_result_batch = list(
        _iter_result_raw(
            dataset,
            model,
            batch_size=BatchSize.LOAD_RECORDS,
            commit=commit,
            include_extracted_properties=include_extracted_properties,
        ),
    )
    return (
        pd.concat(df_result_batch)
        if df_result_batch
        else pd.DataFrame(
            columns=["datapoint_id", COL_DATAPOINT, COL_RESULT, COL_THRESHOLDED_OBJECT, COL_EVAL_CONFIG],
        )
    )


def _process_result(
    eval_config: EvalConfig,
    df_result: pd.DataFrame,
    id_fields: List[str],
    thresholded_fields: Optional[List[str]] = None,
) -> pd.DataFrame:
    df_serialized_datapoint_id_object = _to_serialized_dataframe(
        df_result[sorted(id_fields)],
        column=COL_DATAPOINT_ID_OBJECT,
    )
    if thresholded_fields:
        excludes = id_fields + thresholded_fields
        selected = [col for col in df_result.columns.tolist() if col not in excludes]
        df_result_eval = _to_serialized_dataframe(df_result[selected], column=COL_RESULT)
        df_result_eval[COL_THRESHOLDED_OBJECT] = _to_serialized_dataframe(
            df_result[thresholded_fields],
            column=COL_THRESHOLDED_OBJECT,
        )
    else:
        selected = [col for col in df_result.columns.tolist() if col not in id_fields]
        df_result_eval = _to_serialized_dataframe(df_result[selected], column=COL_RESULT)
    df_result_eval[COL_EVAL_CONFIG] = json.dumps(eval_config) if eval_config is not None else None
    df_result_eval = pd.concat([df_result_eval, df_serialized_datapoint_id_object], axis=1)
    return df_result_eval


def _send_upload_results_request(
    model: str,
    load_uuid: str,
    dataset_id: int,
    sources: Optional[List[Dict[str, str]]],
) -> UploadResultsResponse:
    request = UploadResultsRequest(
        model=model,
        uuid=load_uuid,
        dataset_id=dataset_id,
        sources=sources,
    )
    response = krequests.post(Path.UPLOAD_RESULTS, json=asdict(request))
    krequests.raise_for_status(response)
    return from_dict(UploadResultsResponse, response.json())


@with_event(EventAPI.Event.FETCH_DATASET_MODEL_RESULT)
def download_results(
    dataset: str,
    model: str,
    commit: Optional[str] = None,
    include_extracted_properties: bool = False,
) -> Tuple[pd.DataFrame, List[EvalConfigResults]]:
    """
    Download results given dataset name and model name.

    Concat dataset with results:

    ```python
    df_dp, results = download_results("dataset name", "model name")
    for eval_config, df_result in results:
        df_combined = pd.concat([df_dp, df_result], axis=1)
    ```

    :param dataset: The name of the dataset.
    :param model: The name of the model.
    :param commit: The commit hash for version control. Get the latest commit when this value is `None`.
    :param include_extracted_properties: If True, include kolena extracted properties from automated extractions
    in the datapoints and results as separate columns
    :return: Tuple of DataFrame of datapoints and list of [`EvalConfigResults`][kolena.dataset.EvalConfigResults].
    """
    log.info(f"downloading results for model '{model}' on dataset '{dataset}'")
    existing_dataset = _load_dataset_metadata(dataset)
    assert existing_dataset

    id_fields = existing_dataset.id_fields

    df = _fetch_results(dataset, model, commit, include_extracted_properties)

    if df.empty:
        df_datapoints = pd.DataFrame()
    else:
        df_datapoints = _to_deserialized_dataframe(df.drop_duplicates(subset=[COL_DATAPOINT]), column=COL_DATAPOINT)

    eval_configs = df[COL_EVAL_CONFIG].unique()
    has_thresholded = not df[COL_THRESHOLDED_OBJECT].isnull().all()
    df_results_by_eval = []
    for eval_config in eval_configs:
        df_matched = df[df[COL_EVAL_CONFIG] == eval_config if eval_config is not None else df[COL_EVAL_CONFIG].isnull()]
        df_result = pd.concat(
            [
                _to_deserialized_dataframe(df_matched, column=COL_DATAPOINT)[id_fields],
                _to_deserialized_dataframe(df_matched, column=COL_RESULT),
            ],
            axis=1,
        )
        if has_thresholded:
            df_result = pd.concat(
                [df_result, _to_deserialized_dataframe(df_matched, column=COL_THRESHOLDED_OBJECT)],
                axis=1,
            )
        df_results_by_eval.append(
            EvalConfigResults(json.loads(eval_config) if eval_config is not None else None, df_result),
        )
    log.info(f"downloaded results for model '{model}' on dataset '{dataset}'")
    return df_datapoints, df_results_by_eval


def _validate_configs(configs: List[EvalConfig]) -> None:
    n = len(configs)
    for i in range(n):
        for j in range(i + 1, n):
            if configs[i] == configs[j]:
                raise IncorrectUsageError("duplicate eval configs are invalid")


def _prepare_upload_results_request(
    dataset: str,
    results: Union[DataFrame, List[Tuple[EvalConfig, DataFrame]]],
    thresholded_fields: Optional[List[str]] = None,
) -> Tuple[str, int, int]:
    existing_dataset = _load_dataset_metadata(dataset)
    assert existing_dataset

    id_fields = existing_dataset.id_fields

    if isinstance(results, pd.DataFrame) or isinstance(results, Iterator):
        results = [(None, results)]
    load_uuid = init_upload().uuid

    _validate_configs([cfg for cfg, _ in results])
    total_rows = 0
    for config, df_result_input in results:
        log.info(f"uploading test results with configuration {config}" if config else "uploading test results")
        if isinstance(df_result_input, pd.DataFrame):
            total_rows += df_result_input.shape[0]
            validate_dataframe_ids(df_result_input, id_fields)
            validate_dataframe_columns(df_result_input, id_fields, thresholded_fields)
            df_results = _process_result(config, df_result_input, id_fields, thresholded_fields)
            upload_data_frame(df=df_results, load_uuid=load_uuid)

        else:
            id_column_validated = False
            for df_result in df_result_input:
                if not id_column_validated:
                    validate_dataframe_ids(df_result, id_fields)
                    validate_dataframe_columns(df_result, id_fields, thresholded_fields)
                    id_column_validated = True
                total_rows += df_result.shape[0]
                df_results = _process_result(config, df_result, id_fields, thresholded_fields)
                upload_data_frame(df=df_results, load_uuid=load_uuid)
    dataset_id = existing_dataset.id
    return load_uuid, dataset_id, total_rows


def _upload_results(
    dataset: str,
    model: str,
    results: Union[DataFrame, List[Tuple[EvalConfig, DataFrame]]],
    sources: Optional[List[Dict[str, str]]] = DEFAULT_SOURCES,
    thresholded_fields: Optional[List[str]] = None,
) -> UploadResultsResponse:
    load_uuid, dataset_id, total_rows = _prepare_upload_results_request(dataset, results, thresholded_fields)

    response = _send_upload_results_request(model, load_uuid, dataset_id, sources=sources)
    if isinstance(response.eval_config_id, list):
        models = [serialize_models_url(response.model_id, eval_config_id) for eval_config_id in response.eval_config_id]
    else:
        models = [serialize_models_url(response.model_id, response.eval_config_id)]
    models_str = "&".join([f"models={model}" for model in models])

    link = f"{get_platform_url()}/dataset/standards?datasetId={dataset_id}&{models_str}"
    log.info(
        f"uploaded test results for model '{model}' on dataset '{dataset}': "
        f"{total_rows} uploaded, {response.n_inserted} inserted, {response.n_updated} updated ({link})",
    )
    return response


@with_event(EventAPI.Event.UPLOAD_DATASET_MODEL_RESULT)
def upload_results(
    dataset: str,
    model: str,
    results: Union[DataFrame, List[EvalConfigResults]],
    thresholded_fields: Optional[List[str]] = None,
) -> None:
    """
    This function is used for uploading the results from a specified model on a given dataset.

    :param dataset: The name of the dataset.
    :param model: The name of the model.
    :param results: Either a DataFrame or a list of [`EvalConfigResults`][kolena.dataset.EvalConfigResults].
    :param thresholded_fields: Columns in result DataFrame containing data associated with different thresholds.

    :return: None
    """
    _upload_results(dataset, model, results, thresholded_fields=thresholded_fields)
